var _ = require('underscore'),
    query = {
        indices: {
            relation: ['type', 'from', 'to'],
            asset: ['type', 'isInitial']
        }
    };

query.createValueMatcher = function (value) {
    if (typeof value === 'function') {
        return value;
    } else if (_.isRegExp(value)) {
        return function matchRegExp(obj) {
            return value.test(obj);
        };
    } else if (_.isArray(value)) {
        return function matchArrayMember(obj) {
            return value.indexOf(obj) !== -1;
        };
    } else if (typeof value === 'object' && !value.isAsset && !value.isRelation) {
        return query.queryObjToMatcherFunction(value);
    } else { // Asset or relation or primitive value
        return function matchEqual(obj) {
            return obj === value;
        };
    }
};

query.queryObjToMatcherFunction = function (queryObj) {
    if (typeof queryObj === 'function') {
        return queryObj;
    }
    var matchers = _.map(queryObj || {}, function (value, key) {
        var valueMatcher = query.createValueMatcher(value);
        return function matchPropertyValue(obj) {
            return obj && valueMatcher(obj[key]);
        };
    });
    if (matchers.length === 0) {
        // Make sure that a everything is matched if queryObj is falsy, undefined or {}
        // Amazingly, this is a useful special case.
        return function matchEverything() {
            return true;
        };
    } else if (matchers.length === 1) {
        return matchers[0];
    } else {
        return function (obj) {
            return matchers.every(function (matcher) {
                return matcher(obj);
            });
        };
    }
};

query.createPrefixMatcher = function (prefix) {
    return function (value) {
        return typeof value === 'string' && value.indexOf(prefix) === 0;
    };
};

query.queryAssetGraph = function (assetGraph, objType, queryObj) {
    var numFields = 0,
        filters = [],
        shortestIndexLookup;
    _.each(queryObj || {}, function (fieldQueryObj, fieldName) {
        if (objType === 'asset' && (fieldName === 'incoming' || fieldName === 'outgoing')) {
            filters.push(function (asset) {
                var relationQueryObj = _.clone(fieldQueryObj);
                relationQueryObj[fieldName === 'incoming' ? 'to' : 'from'] = asset;
                return assetGraph.findRelations(relationQueryObj).length > 0;
            });
        } else {
            numFields += 1;
            var indexLookup;
            if (query.indices[objType].indexOf(fieldName) !== -1) {
                if (_.isArray(fieldQueryObj) && fieldQueryObj.every(function (item) {
                    // Don't attempt index lookups if the array contains a matcher function or something even worse:
                    return item.isAsset || item.isRelation || /^(?:string|boolean|number)$/.test(item);
                })) {
                    indexLookup = [];
                    fieldQueryObj.forEach(function (value) {
                        Array.prototype.push.apply(indexLookup, assetGraph._lookupIndex(objType, fieldName, value));
                    });
                } else {
                    var queryType = typeof fieldQueryObj;
                    if (queryType === 'string' || queryType === 'boolean' || queryType === 'number' || fieldQueryObj.isAsset || fieldQueryObj.isRelation) {
                        indexLookup = assetGraph._lookupIndex(objType, fieldName, fieldQueryObj);
                    }
                }
                if (indexLookup && (!shortestIndexLookup || indexLookup.length < shortestIndexLookup.length)) {
                    shortestIndexLookup = indexLookup;
                }
            }
        }
    });
    var result;
    if (shortestIndexLookup && numFields === 1) {
        result = shortestIndexLookup;
    } else {
        result = shortestIndexLookup || assetGraph['_' + objType + 's'];
        var filteredQueryObj = _.clone(queryObj);
        delete filteredQueryObj.incoming;
        delete filteredQueryObj.outgoing;
        filters.unshift(query.queryObjToMatcherFunction(filteredQueryObj));
    }
    if (filters.length === 0) {
        return [].concat(result);
    } else {
        filters.forEach(function (filter) {
            result = result.filter(filter);
        });
        return result;
    }
};

// Helpers that allow queries like:
// assetGraph.findRelations({
//    from: {
//        type: query.not(['Html', 'Css'])
//    },
//    to: {
//        url: query.isDefined
//    }
// });

query.not = function (value) {
    // Warn against subtle mistakes like query.not('HtmlAnchor', 'HtmlImage') which would otherwise silently ignore the second argument.
    if (arguments.length > 1) {
        throw new Error("query.not: More than one argument provided, did you mean to pass an array?");
    }
    var valueMatcher = query.createValueMatcher(value);
    return function (obj) {
        return !valueMatcher(obj);
    };
};

query.or = function () { // ...
    if (arguments.length < 2) {
        throw new Error("query.or: Two or more arguments must be provided");
    }
    var valueMatchers = _.toArray(arguments).map(query.createValueMatcher);
    return function (obj) {
        for (var i = 0 ; i < valueMatchers.length ; i += 1) {
            if (valueMatchers[i](obj)) {
                return true;
            }
        }
        return false;
    };
};

query.and = function () { // ...
    if (arguments.length < 2) {
        throw new Error("query.or: Two or more arguments must be provided");
    }
    var valueMatchers = _.toArray(arguments).map(query.createValueMatcher);
    return function (obj) {
        for (var i = 0 ; i < valueMatchers.length ; i += 1) {
            if (!valueMatchers[i](obj)) {
                return false;
            }
        }
        return true;
    };
};

query.isDefined = function (value) {
    return typeof value !== 'undefined';
};

query.isUndefined = function (value) {
    return typeof value === 'undefined';
};

_.extend(exports, query);
