var seq = require('seq'),
    _ = require('underscore'),
    passError = require('../util/passError'),
    assets = require('../assets'),
    relations = require('../relations'),
    bundleStrategyByName = {};

// Internal helper function. Reuses the parse trees of existing assets, so be careful!
function makeBundle(assetGraph, queryObj, assetsToBundle, cb) {
    if (assetsToBundle.length === 0) {
        return cb(new Error('makeBundle: Bundle must contain at least one asset'));
    } else if (assetsToBundle.length === 1) {
        // Shortcut
        return process.nextTick(function () {
            cb(null, [assetsToBundle[0]]);
        });
    }
    seq(assetsToBundle)
        .parMap(function (asset) {
            asset.getParseTree(this);
        })
        .unflatten()
        .seq(function (parseTrees) {
            var bundleAsset = new assets[queryObj.type]({
                parseTree: assets[queryObj.type].mergeParseTrees(parseTrees)
            });
            bundleAsset.url = assetGraph.root + bundleAsset.id + bundleAsset.getExtension(); // FIXME
            assetGraph.addAsset(bundleAsset);

            var outgoingRelations = assetGraph.findRelations({from: assetsToBundle});
            outgoingRelations.forEach(function (outgoingRelation) {
                assetGraph.removeRelation(outgoingRelation);
                outgoingRelation.from = bundleAsset;
                assetGraph.addRelation(outgoingRelation);
            });

            var seenReferringAssets = {};
            assetGraph.findRelations({type: queryObj.incoming.type, to: assetsToBundle}).forEach(function (incomingRelation) {
                if (!(incomingRelation.from.id in seenReferringAssets)) {
                    assetGraph.attachAndAddRelation(new relations[queryObj.incoming.type]({
                        from: incomingRelation.from,
                        to: bundleAsset
                    }), 'before', incomingRelation);
                    seenReferringAssets[incomingRelation.from.id] = true;
                }
                assetGraph.detachAndRemoveRelation(incomingRelation);
            });

            assetsToBundle.forEach(function (asset) {
                if (assetGraph.findRelations({to: asset}).length === 0) {
                    assetGraph.removeAsset(asset);
                }
            });

            cb(null, bundleAsset);
        })
        ['catch'](cb);
}

// Quick and dirty bundling strategy that gets you down to one <script> and one <link rel='stylesheet'>
// per document, but doesn't do any cross-page optimization.
bundleStrategyByName.oneBundlePerIncludingAsset = function (assetGraph, queryObj, cb) {
    var assetsToBundleById = {},
        seenIncludingAssets = {},
        bundleAssets = [];

    assetGraph.findAssets(queryObj).forEach(function (asset) {
        assetsToBundleById[asset.id] = asset;
        assetGraph.findRelations({type: queryObj.incoming.type, to: asset}).forEach(function (incomingRelation) {
            seenIncludingAssets[incomingRelation.from.id] = incomingRelation.from;
        });
    });

    seq(_.values(seenIncludingAssets))
        .seqEach(function (includingAsset) {
            var callback = this;
            seq(assetGraph.findRelations({from: includingAsset, type: queryObj.incoming.type}).filter(function (relation) {
                return relation.to.id in assetsToBundleById;
            }))
                .parMap(function (outgoingRelation) {
                    if (assetGraph.findRelations({to: outgoingRelation.to}).length > 1) {
                        assetGraph.cloneAsset(outgoingRelation.to, [outgoingRelation], this);
                    } else {
                        this(null, outgoingRelation.to);
                    }
                })
                .unflatten()
                .seq(function (assetsToBundle) {
                    makeBundle(assetGraph, queryObj, assetsToBundle, this);
                })
                .seq(function (bundleAsset) {
                    bundleAssets.push(bundleAsset);
                    callback();
                })
                ['catch'](callback);
        })
        .seq(function () {
            cb(null, bundleAssets);
        })
        ['catch'](cb);
};

// Cross-page optimizing bundling strategy that never puts the same chunk in multiple bundles, but still tries
// to create as few bundles as possible. Also preserves inclusion order.
// FIXME: This bundling strategy is still quite buggy, please don't use it yet.
bundleStrategyByName.sharedBundles = function (assetGraph, queryObj, cb) {
    var allAssets = assetGraph.findAssets(queryObj),
        assetIndex = {},
        seenIncludingAssets = {},
        bundles = [];

    allAssets.forEach(function (asset) {
        assetIndex[asset.id] = null; // Means not in a bundle yet
        assetGraph.findRelations({type: queryObj.incoming.type, to: asset}).forEach(function (incomingRelation) {
            seenIncludingAssets[incomingRelation.from.id] = incomingRelation.from;
        });
    });

    function splitBundle(bundle, index) {
        var newBundle = bundle.splice(index);
        newBundle.forEach(function (asset) {
            assetIndex[asset.id] = newBundle;
        });
        bundles.push(newBundle);
        return newBundle;
    }

    _.values(seenIncludingAssets).forEach(function (includingAsset) {
        var outgoingRelations = assetGraph.findRelations({from: includingAsset, type: queryObj.incoming.type}, true), // includeUnresolved
            previousBundle,
            canAppendToPreviousBundle = false,
            previousBundleIndex;

        outgoingRelations.forEach(function (outgoingRelation) {
            var existingBundle = assetIndex[outgoingRelation.to.id];
            if (existingBundle === null) {
                // Not bundled yet, append to previousBundle if possible, else create a new one
                if (canAppendToPreviousBundle) {
                    previousBundle.push(outgoingRelation.to);
                    previousBundleIndex = previousBundle.length - 1;
                } else {
                    if (previousBundle && previousBundleIndex !== previousBundle.length - 1) {
                        splitBundle(previousBundle, previousBundleIndex);
                    }
                    previousBundle = [outgoingRelation.to];
                    previousBundleIndex = 0;
                    bundles.push(previousBundle);
                    canAppendToPreviousBundle = true;
                }
                assetIndex[outgoingRelation.to.id] = previousBundle;
            } else if (existingBundle) {
                // Already in another bundle
                canAppendToPreviousBundle = false;
                var indexInExistingBundle = existingBundle.indexOf(outgoingRelation.to);
                if (previousBundle && existingBundle === previousBundle) {
                    if (indexInExistingBundle === previousBundleIndex + 1) {
                        previousBundleIndex = indexInExistingBundle;
                    } else {
                        splitBundle(previousBundle, indexInExistingBundle + 1);
                        existingBundle = assetIndex[outgoingRelation.to.id];
                        indexInExistingBundle = existingBundle.indexOf(outgoingRelation.to);
                        if (indexInExistingBundle !== 0) {
                            existingBundle = splitBundle(existingBundle, indexInExistingBundle);
                        }
                        previousBundle = existingBundle;
                        previousBundleIndex = 0;
                    }
                } else {
                    if (previousBundle && previousBundleIndex !== (previousBundle.length - 1)) {
                        splitBundle(previousBundle, previousBundleIndex + 1);
                    }
                    if (indexInExistingBundle !== 0) {
                        existingBundle = splitBundle(existingBundle, indexInExistingBundle);
                    }
                    previousBundle = existingBundle;
                    previousBundleIndex = 0;
                }
            } else {
                // The relation doesn't point at an asset matched by queryObj
                previousBundle = null;
                canAppendToPreviousBundle = false;
            }
        });
        // No more outgoing relations for this asset, make sure that the asset that was bundled
        // last is at the last position in its bundle:
        if (previousBundle && previousBundleIndex !== previousBundle.length - 1) {
            splitBundle(previousBundle, previousBundleIndex + 1);
        }
    });

    seq(bundles)
        .parMap(function (bundle) {
            makeBundle(assetGraph, queryObj, bundle, this);
        })
        .unflatten()
        .seq(function (bundleAssets) {
            cb(null, bundleAssets);
        })
        ['catch'](cb);
};

module.exports = function (queryObj, bundleStrategyName) {
    if (!('type' in queryObj) || !/^(?:Css|JavaScript)$/.test(queryObj.type)) {
        throw new Error("transforms.bundleAssets: queryObj must have a 'type' property of 'JavaScript' or 'Css'");
    }
    if (!queryObj.incoming || !queryObj.incoming.type) {
        throw new Error("transforms.bundleAssets: queryObj must have an 'incoming' obj with a 'type' property");
    }
    if (!bundleStrategyName) {
        bundleStrategyName = 'oneBundlePerIncludingAsset';
    } else if (!(bundleStrategyName in bundleStrategyByName)) {
        throw new Error("transforms.bundleAssets: Unknown bundle strategy: " + bundleStrategyName);
    }

    return function bundleAssets(assetGraph, cb) {
        bundleStrategyByName[bundleStrategyName](assetGraph, queryObj, passError(cb, function (bundleAssets) {
            assetGraph.recomputeBaseAssets();
            bundleAssets.forEach(function (bundleAsset) {
                assetGraph.findRelations({to: bundleAsset}).forEach(function (incomingRelation) {
                    assetGraph.refreshRelationUrl(incomingRelation);
                });
                assetGraph.findRelations({from: bundleAsset}).forEach(function (outgoingRelation) {
                    assetGraph.refreshRelationUrl(outgoingRelation);
                });
            });
            cb();
        }));
    };
};
